/*
 * @(#)Message.java
 *
 * $Revision: 1.5 $
 * $Date: 2006/04/06 19:10:12 $
 *
 * Copyright (c) 2005-2006 RWTH Aachen University, Lehr- und Forschungsgebiet Informatik 4.
 * All rights reserved.
 */

package de.rwth.informatik.lufgi4.trustedpals.messaging;

import java.io.Serializable;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.rwth.informatik.lufgi4.trustedpals.protocol.Item;
import de.rwth.informatik.lufgi4.trustedpals.util.Utils;

/**
 * The message sent between participants.
 * <p>
 * Expected message format: 
 * <pre>
 * +------+--------+---------+-----+---
 * |sender|receiver|hostCount|round|...
 * +------+--------+---------+-----+---
 * </pre>
 * <p>
 * In case round = 0:
 * <pre>
 * -------+------------+---------------------+------------------+-----------------+--------------+
 * round=0|randomNumber|itemDescriptionlength|itemDescription...|itemContentLength|itemContent...|
 * -------+------------+---------------------+------------------+-----------------+--------------+
 * </pre>
 * <p>
 * In case round > 0:
 * <pre>
 * -------+-----+---------------------+-----------------------+
 * round>0|flags|vote(hostCount bytes)|locked(hostCount bytes)|
 * -------+-----+---------------------+-----------------------+
 * </pre>
 * or:
 * <pre>
 * -------+-----+-------------+
 * round>0|flags|value(1 byte)|
 * -------+-----+-------------+
 * </pre>
 * 
 * @author Milan Fort
 * @version $Revision: 1.5 $, $Date: 2006/04/06 19:10:12 $
 * 
 * @since 1.0
 */
public final class Message implements Serializable {
    
    private static final long serialVersionUID = 8351L;
    
    public static final int OFFSET_SENDER = 0;
    
    public static final int OFFSET_RECEIVER = 1;
    
    public static final int OFFSET_HOST_COUNT = 2;
    
    public static final int OFFSET_ROUND = 3;
    
    public static final int OFFSET_RANDOM_NUMBER = 4;
    
    public static final int OFFSET_ITEM = 5;
    
    public static final int OFFSET_FLAGS = 4;
    
    public static final int OFFSET_VOTE = 5;
    
    public static final int FLAG_NORMAL = 1;
    
    public static final int FLAG_FAKE = 2;
    
    public static final int FLAG_DUMMY = 3;

    /* for randomized fair exchange */

    public static final int OFFSET_VALUE = 5;

    public static final int FLAG_PREFERENCE = 4;
    
    public static final int FLAG_DISAGREEMENT = 5;
    
    public static final int FLAG_DECISION = 6;
    
    private final int sender;
    
    private final int receiver;
    
    private final int hostCount;
    
    private final int round;
    
    private int randomNumber = 0;
    
    private Item item = null;
    
    private byte[] vote = null;
    
    private byte[] locked = null;
    
    private byte value = -1;
    
    private boolean normal = false;
    
    private boolean dummy = false;
    
    private boolean fake = false;
    
    private boolean preference = false;
    
    private boolean disagreement = false;
    
    private boolean decision = false;
    
    private final byte[] content;
    
    public Message(byte[] content) {
        this.content = content;
        
        sender = content[OFFSET_SENDER];
        receiver = content[OFFSET_RECEIVER];
        hostCount = content[OFFSET_HOST_COUNT];
        round = content[OFFSET_ROUND];
        
        if (round == 0) {
            randomNumber = content[OFFSET_RANDOM_NUMBER];
            
            int descriptionLength = content[OFFSET_ITEM];
            
            // this message contains no item
            if (descriptionLength == 0) {
                return;
            }
            
            byte[] description = new byte[descriptionLength];
            System.arraycopy(content, OFFSET_ITEM + 1, description, 0, descriptionLength);

            int itemContentLength = content[OFFSET_ITEM + descriptionLength + 1];
            byte[] itemContent = new byte[itemContentLength];
            System.arraycopy(content, OFFSET_ITEM + descriptionLength + 2,
                    itemContent, 0, itemContentLength);
            
            item = new Item(description, itemContent);
            
        } else if (round > 0) {
            if (content[OFFSET_FLAGS] == 0) {
                throw new IllegalArgumentException("Flags field may not be zero");
            }
            
            normal = content[OFFSET_FLAGS] == FLAG_NORMAL;
            fake = content[OFFSET_FLAGS] == FLAG_FAKE;
            dummy = content[OFFSET_FLAGS] == FLAG_DUMMY;

            preference = content[OFFSET_FLAGS] == FLAG_PREFERENCE;
            disagreement = content[OFFSET_FLAGS] == FLAG_DISAGREEMENT;
            decision = content[OFFSET_FLAGS] == FLAG_DECISION;

            if (normal || fake || dummy) {
                vote = new byte[hostCount];
                locked = new byte[hostCount];
                
                if (!fake) {
                    System.arraycopy(content, OFFSET_VOTE, vote, 0, hostCount);
                    System.arraycopy(content, OFFSET_VOTE + hostCount, locked, 0, hostCount);
                } 
                
            } else {
                assert preference || disagreement || decision;
                
                value = content[OFFSET_VALUE];
            }
            
        } else {
            assert round < 0;
            throw new IllegalArgumentException("Round may not be less than zero");
        }
    }
    
    public int getSender() {
        return sender;
    }
    
    public int getReceiver() {
        return receiver;
    }
    
    public int getHostCount() {
        return hostCount;
    }
    
    public int getRound() {
        return round;
    }
    
    public int getRandomNumber() {
        if (round != 0) {
            throw new IllegalStateException("Random number is undefined for this message");
        }
        
        return randomNumber;
    }
    
    public Item getItem() {
        if (round != 0) {
            throw new IllegalStateException("Item is undefined for this message");
        }
        
        return item;
    }
    
    public byte[] getVote() {
        if (round == 0) {
            throw new IllegalStateException("Vote is undefined for this message");
        }
        
        return vote;
    }
    
    public byte[] getLocked() {
        if (round == 0) {
            throw new IllegalStateException("Locked is undefined for this message");
        }
        
        return locked;
    }
    
    public boolean isDummy() {
        if (round == 0) {
            throw new IllegalStateException("Dummy is undefined for this message");
        }
        
        return dummy;
    }
    
    public boolean isFake() {
        if (round == 0) {
            throw new IllegalStateException("Fake is undefined for this message");
        }
        
        return fake;
    }
    
    public byte getValue() {
        if (round == 0) {
            throw new IllegalStateException("Value is undefined for this message");
        }
        
        return value;
    }

    public int getSize() {
        return content.length;
    }
    
    public byte[] getContent() {
        return content;
    }
    
    public boolean equals(Object obj) {
        return EqualsBuilder.reflectionEquals(this, obj);
    }
    
    public int hashCode() {
        return HashCodeBuilder.reflectionHashCode(17, 37, this);
    }
    
    public String toString() {
        return ToStringBuilder.reflectionToString(this, Utils.SHORT_STYLE);
    }
    
    /** Self-testing module. */
    public static void main(String[] args) {
        Message message = new Message(new byte[] {1, 2, 5, 0, 3, 2, 1, 1, 3, 2, 2, 2});
        System.out.println(message);

        message = new Message(new byte[] {1, 2, 3, 1, 3, 1, 1, 1, 0, 0, 0});
        System.out.println(message);
        
        message = new Message(new byte[] {1, 2, 3, 1, 4, 0});
        System.out.println(message);
    }
}
