/*
 * @(#)CSFairExchangeProtocol.java
 *
 * $Revision: 1.2 $
 * $Date: 2006/04/06 19:10:12 $
 *
 * Copyright (c) 2005-2006 RWTH Aachen University, Lehr- und Forschungsgebiet Informatik 4.
 * All rights reserved.
 */

package de.rwth.informatik.lufgi4.trustedpals.protocol;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.rwth.informatik.lufgi4.trustedpals.messaging.Message;
import de.rwth.informatik.lufgi4.trustedpals.random.Random;
import de.rwth.informatik.lufgi4.trustedpals.util.Utils;

/**
 * 
 * 
 * @author Milan Fort
 * @version $Revision: 1.2 $, $Date: 2006/04/06 19:10:12 $
 * 
 * @since 1.0
 */
public final class CSFairExchangeProtocol implements Protocol {
    
    //  possible values are up to 42
    private static final byte MAX_SUPPORTED_HOSTS = 10;
    
    private static final short MAX_SUPPORTED_DATA_LENGTH =
        (MAX_SUPPORTED_HOSTS - 1) * 6 + 1;
    
    private static final byte MAX_ROUNDS = 100;
    
    private static final byte MAX_ITEM_DESCRIPTION_LENGTH = 7;
    
    private static final byte MAX_ITEM_CONTENT_LENGTH = 10;
    
    private static Log logger = LogFactory.getLog(CSFairExchangeProtocol.class);
    
    private byte hostID = 0;
    
    private byte hostCount = 0;
    
    private byte sourceHost = 0;
    
    private byte destinationHost = 0;
    
    private byte[] wantedItemDescription;

    private boolean shouldReturnReceivedItem = false;
    
    private byte receivedItemDescriptionLength = 0;
    
    private byte[] receivedItemDescription = new byte[MAX_ITEM_DESCRIPTION_LENGTH];

    private byte receivedItemContentLength = 0;
    
    private byte[] receivedItemContent = new byte[MAX_ITEM_CONTENT_LENGTH];
    
    private Item ownItem;
    
    /** Data received for the current's round computation. */
    private byte[] data = null;
    
    /** Index to data array, where additional data from the next subround should be appended. */
    private short dataIndex = 0;
    
    /** Messages to be sent out during the next sending phase. */
    private byte[] messages = null;
    
    /** Current round. */
    private byte round = 0;
    
    /** Random number k. */
    private byte fakeRoundsCount = 0;
    
    private volatile boolean finished = false;
    
    private volatile boolean computationFinished = false;

    private byte[] seed = null;

    private boolean[] disagreementReceivedInRound = null;
    
    private byte initialProposition = 0;
    
    private byte subround = 0;
    
    private Random random = null;
    
    public CSFairExchangeProtocol() {
        assert MAX_SUPPORTED_DATA_LENGTH < 250;
        
        messages = new byte[MAX_SUPPORTED_DATA_LENGTH];
        data = new byte[MAX_SUPPORTED_DATA_LENGTH];
        
        disagreementReceivedInRound = new boolean[MAX_ROUNDS];
        
        random = new Random();
    }

    private void clearMessages() {
        Arrays.fill(messages, (byte) 0);
    }
    
    private void clearData() {
        dataIndex = 0;
        Arrays.fill(data, (byte) 0);
    }
    
    private void clearDisagreementWatcher() {
        Arrays.fill(disagreementReceivedInRound, false);
    }
    
    private void clearReceivedItem() {
        receivedItemDescriptionLength = 0;
        receivedItemContentLength = 0;
        
        Arrays.fill(receivedItemDescription, (byte) 0);
        Arrays.fill(receivedItemContent, (byte) 0);
    }
    
    public List getMessages() {
//        if (!computationFinished) {
//            throw new IllegalStateException("Computation has not finished yet");
//        }
        
        List parsedMessages = new ArrayList();
        byte currentPosition = 0;

        while (currentPosition < messages.length) {
            byte sender = messages[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            byte hostCount = messages[currentPosition + Message.OFFSET_HOST_COUNT];
            byte round = messages[currentPosition + Message.OFFSET_ROUND];

            byte[] content = null;
            if (round == 0) {
                byte descriptionLength = messages[currentPosition + Message.OFFSET_ITEM];
                
                if (descriptionLength == 0) {
                    content = new byte[6];
                    
                } else {
                    byte itemContentLength = messages[currentPosition + Message.OFFSET_ITEM
                                                     + descriptionLength + 1];
                    
                    content = new byte[descriptionLength + itemContentLength + 2 + 5];
                }
                
            } else {
                assert hostCount == this.hostCount;
                
                content = new byte[normalMessageLength()];
            }
            
            System.arraycopy(messages, currentPosition, content, 0, content.length);
            Message message = new Message(content);
            parsedMessages.add(message);
            currentPosition += content.length;
        }
        
        return parsedMessages;
    }
    
    boolean isInitiator() {
        if (hostCount <= 0) {
            throw new IllegalStateException("Host count must be greater then zero");
        }
        
        if (hostID <= 0) {
            throw new IllegalStateException("Host ID must be greater then zero");
        }
        
        return hostID == 1;
    }
    
    byte pickRandomNumber() {
        byte randomNumber = (byte) (Math.random() * 10);
        
        randomNumber %= 4;
        randomNumber++;
        
        return randomNumber;
    }
    
    /**
     * 
     * 
     * @return the length of the just prepared message
     */
    byte prepareMessageWithItem() {
        logger.debug("Preparing message with item for host " + destinationHost);
        
        if (isInitiator()) {
            logger.debug("Random number k=" + fakeRoundsCount + " will be also included");
        }
        
        assert round == 0;
        clearMessages();

        messages[Message.OFFSET_SENDER] = (byte) hostID;
        messages[Message.OFFSET_RECEIVER] = (byte) destinationHost;
        messages[Message.OFFSET_HOST_COUNT] = (byte) hostCount;
        messages[Message.OFFSET_ROUND] = (byte) round;
        messages[Message.OFFSET_RANDOM_NUMBER] = (byte) fakeRoundsCount;
        
        byte[] description = ownItem.getDescription();
        byte[] content = ownItem.getContent();
        
        messages[Message.OFFSET_ITEM] = (byte) description.length;
        System.arraycopy(description, 0, messages, Message.OFFSET_ITEM + 1, description.length);
        
        messages[Message.OFFSET_ITEM + description.length + 1] = (byte) content.length;
        System.arraycopy(content, 0, messages, Message.OFFSET_ITEM + description.length + 2,
                content.length);
        
        return  (byte) (Message.OFFSET_ITEM + description.length + content.length + 2);
    }
    
    /**
     * Returns <code>true</true> if an item (correct or not) has been received from source host.
     * <p> 
     * WARNING: This method has side effects. It parses out the received item.
     * 
     * @return <code>true</true> if an item (correct or not) has been received from source host.
     */
    boolean hasReceivedItemFromSourceHost() {
        clearReceivedItem();
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (sender == sourceHost) {
                receivedItemDescriptionLength = data[currentPosition + Message.OFFSET_ITEM];
                
                if (receivedItemDescriptionLength == 0) {
                    return false;
                }
                
                System.arraycopy(data, currentPosition + Message.OFFSET_ITEM + 1,
                        receivedItemDescription, 0, receivedItemDescriptionLength);
                
                receivedItemContentLength = data[currentPosition + Message.OFFSET_ITEM
                                                 + receivedItemDescriptionLength + 1];
                
                System.arraycopy(data, currentPosition + Message.OFFSET_ITEM
                        + receivedItemDescriptionLength + 2, receivedItemContent, 0,
                        receivedItemContentLength);
                
                return true;
                
            } else {
                byte descriptionLength = data[currentPosition + Message.OFFSET_ITEM];
                
                if (descriptionLength == 0) {
                    currentPosition += Message.OFFSET_ITEM + 1;
                    
                } else {
                    assert descriptionLength > 0;
                    
                    byte itemContentLength = data[currentPosition + Message.OFFSET_ITEM
                                                     + descriptionLength + 1];
                    
                    currentPosition += Message.OFFSET_ITEM + descriptionLength + itemContentLength + 2;
                }
            }
        }
        
        return false;
    }
    
    boolean isReceivedItemCorrect() {
        
        if (receivedItemDescriptionLength == wantedItemDescription.length) {
            for (byte i = 0; i < receivedItemDescriptionLength; i++) {
                if (receivedItemDescription[i] != wantedItemDescription[i]) {
                    return false;
                }
            }
            
            for (byte i = receivedItemDescriptionLength; i < receivedItemDescription.length; i++) {
                if (receivedItemDescription[i] != 0) {
                    return false;
                }
            }
            
        } else {
            return false;
        }
        
        return true;
    }
    
    void prepareInitiatorMessages() {
        assert round == 0;
        clearMessages();
        
        byte currentPosition = prepareMessageWithItem();
        
        for (byte i = 1; i <= hostCount; i++) {
            if (i != hostID && i != destinationHost) {
                logger.debug("Preparing message with random number k=" + fakeRoundsCount
                        + " for host " + i);
                
                messages[currentPosition + Message.OFFSET_SENDER] = (byte) hostID;
                messages[currentPosition + Message.OFFSET_RECEIVER] = (byte) i;
                messages[currentPosition + Message.OFFSET_HOST_COUNT] = (byte) hostCount;
                messages[currentPosition + Message.OFFSET_ROUND] = (byte) round;
                messages[currentPosition + Message.OFFSET_RANDOM_NUMBER] = (byte) fakeRoundsCount;
                messages[currentPosition + Message.OFFSET_ITEM] = (byte) 0x0;
                
                currentPosition += 6;
            }
        }
    }
    
    /**
     * Returns <code>true</true> if a random number has been received from initiator.
     * <p> 
     * WARNING: This method has side effects. It parses out the received random number.
     * 
     * @return <code>true</true> if a random number has been received from initiator.
     */
    boolean hasReceivedRandomNumberFromInitiator() {
        if (hostID == 1) {
            return true;
        }
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (sender == 1) {
                fakeRoundsCount = data[currentPosition + Message.OFFSET_RANDOM_NUMBER]; 
                
                return true;
                
            } else {
                byte descriptionLength = data[currentPosition + Message.OFFSET_ITEM];
                
                if (descriptionLength == 0) {
                    currentPosition += Message.OFFSET_ITEM + 1;
                    
                } else {
                    assert descriptionLength > 0;
                    
                    byte itemContentLength = data[currentPosition + Message.OFFSET_ITEM
                                                     + descriptionLength + 1];
                    
                    currentPosition += Message.OFFSET_ITEM + descriptionLength + itemContentLength + 2;
                }
            }
        }
        
        return false;
    }
    
    void prepareFakeMessages() {
        assert round > 0;
        clearMessages();
        
        byte currentPosition = 0;
        
        for (byte i = 1; i <= hostCount; i++) {
            if (i != hostID) {
                messages[currentPosition + Message.OFFSET_SENDER] = (byte) hostID;
                messages[currentPosition + Message.OFFSET_RECEIVER] = (byte) i;
                messages[currentPosition + Message.OFFSET_HOST_COUNT] = (byte) hostCount;
                messages[currentPosition + Message.OFFSET_ROUND] = (byte) round;
                messages[currentPosition + Message.OFFSET_FLAGS] = Message.FLAG_FAKE;
                
                currentPosition += normalMessageLength();
            }
        }
    }
    
    boolean hasReceivedFakeMessagesFromAllHosts() {
        byte[] received = new byte[hostCount];
        received[hostID-1] = 1;
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            received[sender-1] = 1;
            
            currentPosition += normalMessageLength();
        }
        
        for (byte i = 0; i < hostCount; i++) {
            if (received[i] != 1) {
                return false;
            }
        }
        
        return true;
    }
    
    private void prepareMessages(byte flag, byte value) {
        assert round > 0;
        assert value == 0 || value == 1;
        
        clearMessages();
        
        byte currentPosition = 0;
        
        for (byte i = 1; i <= hostCount; i++) {
            if (i != hostID) {
                messages[currentPosition + Message.OFFSET_SENDER] = hostID;
                messages[currentPosition + Message.OFFSET_RECEIVER] = i;
                messages[currentPosition + Message.OFFSET_HOST_COUNT] = hostCount;
                messages[currentPosition + Message.OFFSET_ROUND] = round;
                messages[currentPosition + Message.OFFSET_FLAGS] = flag;
                messages[currentPosition + Message.OFFSET_VALUE] = value;
                
                currentPosition += normalMessageLength();
            }
        }
    }
    
    void prepareDisagreementMessages() {
        
        //Note: value is not important in this type of message; should be set to zero
        prepareMessages((byte) Message.FLAG_DISAGREEMENT, (byte) 0);
    }
    
    void preparePreferenceMessages(byte value) {
        prepareMessages((byte) Message.FLAG_PREFERENCE, value);
    }
    
    void prepareDecisionMessages(byte value) {
        prepareMessages((byte) Message.FLAG_DECISION, value);
    }
    
    private boolean hasReceivedMessage(byte flag) {
        boolean result = false;
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (data[currentPosition + Message.OFFSET_FLAGS] == flag) {
                result = true;
                break;
            }
            
            currentPosition += normalMessageLength();
        }
        
        return result;
    }

    boolean hasReceivedPreferenceMessage() {
        return hasReceivedMessage((byte) Message.FLAG_PREFERENCE);
    }
    
    boolean hasReceivedDisagreementMessage() {
        return hasReceivedMessage((byte) Message.FLAG_DISAGREEMENT);
    }

    boolean hasReceivedDecisionMessage() {
        return hasReceivedMessage((byte) Message.FLAG_DECISION);
    }
    
    boolean hasReceivedBothPreferenceValues() {
        boolean receivedZero = false;
        boolean receivedOne = false;
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (data[currentPosition + Message.OFFSET_FLAGS] == Message.FLAG_PREFERENCE) {
                if (data[currentPosition + Message.OFFSET_VALUE] == 0) {
                    receivedZero = true;
                    
                } else {
                    receivedOne = true;
                }
            }
            
            currentPosition += normalMessageLength();
        }
        
        return receivedZero && receivedOne;
    }
    
    private byte getFirstReceivedValue(byte flag) {
        byte value = -1;
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (data[currentPosition + Message.OFFSET_FLAGS] == flag) {
                value = data[currentPosition + Message.OFFSET_VALUE];
                break;
            }
            
            currentPosition += normalMessageLength();
        }

        return value;
    }
    
    byte getFirstReceivedPreferenceValue() {
        return getFirstReceivedValue((byte) Message.FLAG_PREFERENCE);
    }

    byte getFirstReceivedDecisionValue() {
        return getFirstReceivedValue((byte) Message.FLAG_DECISION);
    }

    public Item getItem() {
        if (finished && shouldReturnReceivedItem) {
            byte[] description = new byte[receivedItemDescriptionLength];
            System.arraycopy(receivedItemDescription, 0, description, 0, description.length);
            
            byte[] content = new byte[receivedItemContentLength];
            System.arraycopy(receivedItemContent, 0, content, 0, content.length);
            
            return new Item(description, content);
                
        } else {
            return new Item();
        }
    }
    
    public boolean hasErrorOccured() {

        // not used yet
        return false;
    }
    
    public boolean isComputationFinished() {
        return computationFinished;
    }
    
    public boolean isFinished() {
        return finished;
    }
    
    public void start() {
        shouldReturnReceivedItem = false;
        computationFinished = false;
        finished = false;
        
        round = 0;
        subround = 1;
        clearData();
        clearDisagreementWatcher();
        clearMessages();
        clearReceivedItem();
        
        // If nothing goes wrong, we want to propose 1
        initialProposition = 1;
        
        logger.info("Starting round " + round);

        if (isInitiator()) {
            logger.trace("I am the Initiator");
            
            byte k = pickRandomNumber();
            logger.debug("Generated random number k=" + k);
            
            fakeRoundsCount = k;
            
            prepareInitiatorMessages();
            
        } else {
            logger.trace("I am NOT the Initiator");
            
            // to be received from initiator later
            fakeRoundsCount = 0;
            
            prepareMessageWithItem();
        }
        
        round++;
        computationFinished = true;
    }
    
    byte[] messagesToBytes(List messages) {
        byte[] temp = new byte[MAX_SUPPORTED_DATA_LENGTH];
        byte currentPosition = 0;
        
        for (Iterator i = messages.iterator(); i.hasNext(); ) {
            Message message = (Message) i.next();
            System.arraycopy(message.getContent(), 0, temp, currentPosition, message.getSize());
            
            currentPosition += message.getSize();
        }
        
        byte[] result = new byte[currentPosition];
        System.arraycopy(temp, 0, result, 0, result.length);
        
        return result;
    }
    
    public void startNextRound(List messages) {
        if (finished) {
            throw new IllegalStateException("Protocol already finished");
        }
        
        computationFinished = false;

        byte[] currentData = messagesToBytes(messages);
        System.arraycopy(currentData, 0, data, dataIndex, currentData.length);
        dataIndex += currentData.length;
        
        clearMessages();
        
        byte randomNumber = random.nextRandomNumber();
        assert randomNumber == 0 || randomNumber == 1;
        
        logger.info("Starting round " + round + ", subround " + subround);
        logger.debug("Current random number is " + randomNumber);
        
        if (round == 1) {
            if (hasReceivedItemFromSourceHost()) {
                logger.debug("Received an item from source host");
                
                if (isReceivedItemCorrect()) {
                    logger.debug("Received item is correct");
                    
                    initialProposition = 1;
                    
                } else {
                    logger.error("Received item is NOT correct");
                    
                    initialProposition = 0;
                }
                
            } else {
                logger.error("Did not receive the item from source host");
                
                initialProposition = 0;
            }
            
            if (!isInitiator()) {
                if (hasReceivedRandomNumberFromInitiator()) {
                    logger.debug("Received fake rounds count: " + fakeRoundsCount);
                    
                } else {
                    logger.error("Did not receive fake rounds count");
                    
                    initialProposition = 0;
                }
            }
            
            prepareFakeMessages();
            clearData();
            round++;
            
        } else if (round <= fakeRoundsCount) {
            logger.debug("This is a fake round");
            
            if (hasReceivedFakeMessagesFromAllHosts()) {
                logger.info("Fake messages received from all hosts");
                
            } else {
                logger.error("Did not receive fake messages from all other hosts");
                
                initialProposition = 0;
            }
            
            prepareFakeMessages();
            clearData();
            round++;
            
        } else if (round == fakeRoundsCount + 1) {
            if (hasReceivedFakeMessagesFromAllHosts()) {
                logger.info("Fake messages received from all hosts");
                
            } else {
                logger.error("Did not receive fake messages from all hosts");
                
                initialProposition = 0;
            }
            
            // do not prepare fake messages anymore
            
            logger.info("Real calculation is starting now");
            logger.debug("Initial proposition is " + initialProposition);
            
            // Figure 3, line 1
            preparePreferenceMessages(initialProposition);
            clearData();
            round++;
            subround = 1;
            
        } else {
            assert round > fakeRoundsCount + 1;
            
            if (subround == 1) {
                if (hasReceivedBothPreferenceValues()) {
                    logger.debug("I have received both preference values");
                    
                    prepareDisagreementMessages();
                    disagreementReceivedInRound[round] = true;
                    
                } else {
                    logger.debug("I have NOT received both preference values");
                }
                
                subround++;
                
            } else if (subround <= hostCount) {
                if (hasReceivedDisagreementMessage() && !disagreementReceivedInRound[round]) {
                    logger.debug("I have received disagreement for the first time");
                    
                    prepareDisagreementMessages();
                    disagreementReceivedInRound[round] = true;
                    
                } else {
                    logger.debug("I have NOT received disagreement for the first time");
                }
                
                subround++;
                
            } else {
                assert subround == hostCount + 1;
                
                if (hasReceivedPreferenceMessage() && !hasReceivedBothPreferenceValues()
                        && !hasReceivedDisagreementMessage()) {
                    
                    logger.debug("I have received only one preference value and no disagreement");
                    
                    byte receivedPreference = getFirstReceivedPreferenceValue();
                    assert receivedPreference == 0 || receivedPreference == 1;
                    
                    logger.debug("Preference value is " + receivedPreference
                            + ", random number is " + randomNumber);
                    
                    if (randomNumber == receivedPreference) {
                        logger.info("I have reached the decision");
                        
                        prepareDecisionMessages(receivedPreference);
                        
                        shouldReturnReceivedItem = receivedPreference == 1;
                        finished = true;
                        
                    } else {
                        logger.debug("Decision not reached yet");
                        
                        preparePreferenceMessages(receivedPreference);
                    }
                    
                } else {
                    logger.debug("I have received a disagreement");
                    
                    preparePreferenceMessages(randomNumber);
                }
                
                if (hasReceivedDecisionMessage()) {
                    byte receivedDecision = getFirstReceivedDecisionValue();
                    assert receivedDecision == 0 || receivedDecision == 1;
                    
                    logger.info("I have received a decision: " + receivedDecision);
                    
                    shouldReturnReceivedItem = receivedDecision == 1;
                    finished = true;
                }
                
                clearData();
                round++;
                subround = 1;
            }
        }
        
        computationFinished = true;
    }
    
    private byte normalMessageLength() {
        return (byte) (1 + 5);
    }
    
    /* Properties */
    
    public void setHostCount(byte hostCount) {
        this.hostCount = hostCount;
    }
    
    public void setHostID(byte hostID) {
        this.hostID = hostID;
    }
    
    public void setDestinationHost(byte destinationHost) {
        this.destinationHost = destinationHost;
    }
    
    public void setSourceHost(byte sourceHost) {
        this.sourceHost = sourceHost;
    }
    
    public void setOwnItem(Item ownItem) {
        this.ownItem = ownItem;
    }
    
    public void setWantedItemDescription(byte[] wantedItemDescription) {
        this.wantedItemDescription = wantedItemDescription;
        
        //TODO: temporary solution to overcome configuration conversion problems
        Utils.convertNumbers(this.wantedItemDescription);
    }

    public void setSeed(byte[] seed) {
        this.seed = seed;
        
        //TODO: temporary solution to overcome configuration conversion problems
        Utils.convertNumbers(this.seed);
        
        random.setSeed(seed, (byte) 0, (byte) seed.length);
    }

    /* Properties not to be set via IoC */
    
    void setMessages(byte[] messages) {
        this.messages = messages;
    }
    
    void setRound(byte round) {
        this.round = round;
    }
    
    void setFakeRoundsCount(byte fakeRoundsCount) {
        this.fakeRoundsCount = fakeRoundsCount;
    }
    
    byte[] getData() {
        return data;
    }
    
    void setData(byte[] data) {
//      this.data = data;
        
        clearData();
        
        System.arraycopy(data, 0, this.data, 0, data.length);
    }
    
    byte getFakeRoundsCount() {
        return fakeRoundsCount;
    }
    
    byte[] getReceivedItemContent() {
        return receivedItemContent;
    }
    
    byte getReceivedItemContentLength() {
        return receivedItemContentLength;
    }
    
    byte[] getReceivedItemDescription() {
        return receivedItemDescription;
    }
    
    byte getReceivedItemDescriptionLength() {
        return receivedItemDescriptionLength;
    }
    
    void setReceivedItemDescription(byte[] receivedItemDescription) {
        this.receivedItemDescription = receivedItemDescription;
    }
    
    void setReceivedItemDescriptionLength(byte receivedItemDescriptionLength) {
        this.receivedItemDescriptionLength = receivedItemDescriptionLength;
    }
}
