/*
 * @(#)GDFairExchangeProtocol.java
 *
 * $Revision: 1.2 $
 * $Date: 2006/04/06 19:10:12 $
 *
 * Copyright (c) 2005-2006 RWTH Aachen University, Lehr- und Forschungsgebiet Informatik 4.
 * All rights reserved.
 */

package de.rwth.informatik.lufgi4.trustedpals.protocol;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.rwth.informatik.lufgi4.trustedpals.messaging.Message;
import de.rwth.informatik.lufgi4.trustedpals.util.Utils;

/**
 * 
 * 
 * @author Milan Fort
 * @version $Revision: 1.2 $, $Date: 2006/04/06 19:10:12 $
 * 
 * @since 1.0
 */
public final class GDFairExchangeProtocol implements Protocol {
    
    // possible values are up to 10
    private static final byte MAX_SUPPORTED_HOSTS = 7;
    
    private static final short MAX_SUPPORTED_DATA_LENGTH =
        (MAX_SUPPORTED_HOSTS - 1) * (5 + (2 * MAX_SUPPORTED_HOSTS)) + 1;
    
    private static final byte MAX_ITEM_DESCRIPTION_LENGTH = 7;
    
    private static final byte MAX_ITEM_CONTENT_LENGTH = 10;
    
    private static Log logger = LogFactory.getLog(GDFairExchangeProtocol.class);
    
    private byte hostID = 0;
    
    private byte hostCount = 0;
    
    private byte sourceHost = 0;
    
    private byte destinationHost = 0;
    
    private byte[] wantedItemDescription;

    private boolean shouldReturnReceivedItem = false;
    
    private byte receivedItemDescriptionLength = 0;
    
    private byte[] receivedItemDescription = new byte[MAX_ITEM_DESCRIPTION_LENGTH];

    private byte receivedItemContentLength = 0;
    
    private byte[] receivedItemContent = new byte[MAX_ITEM_CONTENT_LENGTH];
    
    private HostSet vote;
    
    private HostSet newVote;
    
    private HostSet locked;
    
    private HostSet suspected;
    
    private byte[] receivedVote = new byte[MAX_SUPPORTED_HOSTS];
    
    private byte[] receivedLocked = new byte[MAX_SUPPORTED_HOSTS];
    
    private byte maxFaultyProcesses;
    
    private Item ownItem;
    
    /** Data received for the current's round computation. */
    private byte[] data = null;
    
    /** Messages to be sent out during the next sending phase. */
    private byte[] messages = null;
    
    /** Current round. */
    private byte round = 0;
    
    /** Random number k. */
    private byte fakeRoundsCount = 0;
    
    private volatile boolean finished = false;
    
    private volatile boolean computationFinished = false;
    
    public GDFairExchangeProtocol() {
        assert MAX_SUPPORTED_DATA_LENGTH < 250;
        
        messages = new byte[MAX_SUPPORTED_DATA_LENGTH];
        data = new byte[MAX_SUPPORTED_DATA_LENGTH];
        
        vote = new HostSet(MAX_SUPPORTED_HOSTS);
        newVote = new HostSet(MAX_SUPPORTED_HOSTS);
        locked = new HostSet(MAX_SUPPORTED_HOSTS);
        suspected = new HostSet(MAX_SUPPORTED_HOSTS);
    }

    private void clearMessages() {
        Arrays.fill(messages, (byte) 0);
    }
    
    private void clearData() {
        Arrays.fill(data, (byte) 0);
    }
    
    private void clearReceivedItem() {
        receivedItemDescriptionLength = 0;
        receivedItemContentLength = 0;
        
        Arrays.fill(receivedItemDescription, (byte) 0);
        Arrays.fill(receivedItemContent, (byte) 0);
    }
    
    private void clearReceivedVoteAndLocked() {
        Arrays.fill(receivedVote, (byte) 0);
        Arrays.fill(receivedLocked, (byte) 0);
    }

    public List getMessages() {
        //TODO: check whether the computation is already finished
        
        List parsedMessages = new ArrayList();
        byte currentPosition = 0;

        while (currentPosition < messages.length) {
            byte sender = messages[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            byte hostCount = messages[currentPosition + Message.OFFSET_HOST_COUNT];
            byte round = messages[currentPosition + Message.OFFSET_ROUND];

            byte[] content = null;
            if (round == 0) {
                byte descriptionLength = messages[currentPosition + Message.OFFSET_ITEM];
                
                if (descriptionLength == 0) {
                    content = new byte[6];
                    
                } else {
                    byte itemContentLength = messages[currentPosition + Message.OFFSET_ITEM
                                                     + descriptionLength + 1];
                    
                    content = new byte[descriptionLength + itemContentLength + 2 + 5];
                }
                
            } else {
                assert hostCount == this.hostCount;
                
                /*
                 * Note: We use the instance variable hostCount to determine the expected length
                 * of the message, not the currently parsed one.
                 */
                content = new byte[normalMessageLength()];
            }
            
            System.arraycopy(messages, currentPosition, content, 0, content.length);
            Message message = new Message(content);
            parsedMessages.add(message);
            currentPosition += content.length;
        }
        
        return parsedMessages;
    }
    
    boolean isInitiator() {
        if (hostCount <= 0) {
            throw new IllegalStateException("Host count must be greater then zero");
        }
        
        if (hostID <= 0) {
            throw new IllegalStateException("Host ID must be greater then zero");
        }
        
        return hostID == 1;
    }
    
    byte pickRandomNumber() {
        byte randomNumber = (byte) (Math.random() * 10);
        
        randomNumber %= 4;
        randomNumber++;
        
        return randomNumber;
    }
    
    byte computeMaxFaultyProcesses() {
        return (byte) (hostCount % 2 == 0 ? hostCount / 2 - 1 : hostCount / 2);
    }
    
    /**
     * 
     * 
     * @return the length of the just prepared message
     */
    byte prepareMessageWithItem() {
        logger.debug("Preparing message with item for host " + destinationHost);
        
        if (isInitiator()) {
            logger.debug("Random number k=" + fakeRoundsCount + " will be also included");
        }
        
        assert round == 0;
        clearMessages();

        messages[Message.OFFSET_SENDER] = (byte) hostID;
        messages[Message.OFFSET_RECEIVER] = (byte) destinationHost;
        messages[Message.OFFSET_HOST_COUNT] = (byte) hostCount;
        messages[Message.OFFSET_ROUND] = (byte) round;
        messages[Message.OFFSET_RANDOM_NUMBER] = (byte) fakeRoundsCount;
        
        byte[] description = ownItem.getDescription();
        byte[] content = ownItem.getContent();
        
        messages[Message.OFFSET_ITEM] = (byte) description.length;
        System.arraycopy(description, 0, messages, Message.OFFSET_ITEM + 1, description.length);
        
        messages[Message.OFFSET_ITEM + description.length + 1] = (byte) content.length;
        System.arraycopy(content, 0, messages, Message.OFFSET_ITEM + description.length + 2,
                content.length);
        
        return  (byte) (Message.OFFSET_ITEM + description.length + content.length + 2);
    }
    
    /**
     * Returns <code>true</true> if an item (correct or not) has been received from source host.
     * <p> 
     * WARNING: This method has side effects. It parses out the received item.
     * 
     * @return <code>true</true> if an item (correct or not) has been received from source host.
     */
    boolean hasReceivedItemFromSourceHost() {
        clearReceivedItem();
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (sender == sourceHost) {
                receivedItemDescriptionLength = data[currentPosition + Message.OFFSET_ITEM];
                
                if (receivedItemDescriptionLength == 0) {
                    return false;
                }
                
                System.arraycopy(data, currentPosition + Message.OFFSET_ITEM + 1,
                        receivedItemDescription, 0, receivedItemDescriptionLength);
                
                receivedItemContentLength = data[currentPosition + Message.OFFSET_ITEM
                                                 + receivedItemDescriptionLength + 1];
                
                System.arraycopy(data, currentPosition + Message.OFFSET_ITEM
                        + receivedItemDescriptionLength + 2, receivedItemContent, 0,
                        receivedItemContentLength);
                
                return true;
                
            } else {
                byte descriptionLength = data[currentPosition + Message.OFFSET_ITEM];
                
                if (descriptionLength == 0) {
                    currentPosition += Message.OFFSET_ITEM + 1;
                    
                } else {
                    assert descriptionLength > 0;
                    
                    byte itemContentLength = data[currentPosition + Message.OFFSET_ITEM
                                                     + descriptionLength + 1];
                    
                    currentPosition += Message.OFFSET_ITEM + descriptionLength + itemContentLength + 2;
                }
            }
        }
        
        return false;
    }
    
    boolean isReceivedItemCorrect() {
        
        if (receivedItemDescriptionLength == wantedItemDescription.length) {
            for (byte i = 0; i < receivedItemDescriptionLength; i++) {
                if (receivedItemDescription[i] != wantedItemDescription[i]) {
                    return false;
                }
            }
            
            for (byte i = receivedItemDescriptionLength; i < receivedItemDescription.length; i++) {
                if (receivedItemDescription[i] != 0) {
                    return false;
                }
            }
            
        } else {
            return false;
        }
        
        return true;
    }
    
    void prepareInitiatorMessages() {
        assert round == 0;
        clearMessages();
        
        byte currentPosition = prepareMessageWithItem();
        
        for (byte i = 1; i <= hostCount; i++) {
            if (i != hostID && i != destinationHost) {
                logger.debug("Preparing message with random number k=" + fakeRoundsCount
                        + " for host " + i);
                
                messages[currentPosition + Message.OFFSET_SENDER] = (byte) hostID;
                messages[currentPosition + Message.OFFSET_RECEIVER] = (byte) i;
                messages[currentPosition + Message.OFFSET_HOST_COUNT] = (byte) hostCount;
                messages[currentPosition + Message.OFFSET_ROUND] = (byte) round;
                messages[currentPosition + Message.OFFSET_RANDOM_NUMBER] = (byte) fakeRoundsCount;
                messages[currentPosition + Message.OFFSET_ITEM] = (byte) 0x0;
                
                currentPosition += 6;
            }
        }
    }
    
    boolean hasReceivedRandomNumberFromInitiator() {
        if (hostID == 1) {
            return true;
        }
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (sender == 1) {
                fakeRoundsCount = data[currentPosition + Message.OFFSET_RANDOM_NUMBER]; 
                
                return true;
                
            } else {
                byte descriptionLength = data[currentPosition + Message.OFFSET_ITEM];
                
                if (descriptionLength == 0) {
                    currentPosition += Message.OFFSET_ITEM + 1;
                    
                } else {
                    assert descriptionLength > 0;
                    
                    byte itemContentLength = data[currentPosition + Message.OFFSET_ITEM
                                                     + descriptionLength + 1];
                    
                    currentPosition += Message.OFFSET_ITEM + descriptionLength + itemContentLength + 2;
                }
            }
        }
        
        return false;
    }
    
    void prepareFakeMessages() {
        assert round > 0;
        clearMessages();
        
        byte currentPosition = 0;
        
        for (byte i = 1; i <= hostCount; i++) {
            if (i != hostID) {
                messages[currentPosition + Message.OFFSET_SENDER] = (byte) hostID;
                messages[currentPosition + Message.OFFSET_RECEIVER] = (byte) i;
                messages[currentPosition + Message.OFFSET_HOST_COUNT] = (byte) hostCount;
                messages[currentPosition + Message.OFFSET_ROUND] = (byte) round;
                messages[currentPosition + Message.OFFSET_FLAGS] = Message.FLAG_FAKE;
                
                currentPosition += normalMessageLength();
            }
        }
    }
    
    boolean hasReceivedFakeMessagesFromAllHosts() {
        byte[] received = new byte[hostCount];
        received[hostID-1] = 1;
        
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            received[sender-1] = 1;
            
            currentPosition += normalMessageLength();
        }
        
        for (byte i = 0; i < hostCount; i++) {
            if (received[i] != 1) {
                return false;
            }
        }
        
        return true;
    }
    
    void prepareNormalMessages() {
        assert round > 0;
        clearMessages();
        
        byte currentPosition = 0;
        
        for (byte i = 1; i <= hostCount; i++) {
            if (i != hostID) {
                messages[currentPosition + Message.OFFSET_SENDER] = (byte) hostID;
                messages[currentPosition + Message.OFFSET_RECEIVER] = i;
                messages[currentPosition + Message.OFFSET_HOST_COUNT] = (byte) hostCount;
                messages[currentPosition + Message.OFFSET_ROUND] = (byte) round;
                
                if (suspected.contains(i)) {
                    messages[currentPosition + Message.OFFSET_FLAGS] = Message.FLAG_DUMMY;
                    
                } else {
                    messages[currentPosition + Message.OFFSET_FLAGS] = Message.FLAG_NORMAL;
                }
                
                System.arraycopy(newVote.toArray(), 0, messages,
                        currentPosition + Message.OFFSET_VOTE, hostCount);
                
                System.arraycopy(locked.toArray(), 0, messages,
                        currentPosition + Message.OFFSET_VOTE + hostCount, hostCount);
                
                currentPosition += normalMessageLength();
            }
        }
    }
    
    boolean hasReceivedNormalMessageFromHost(byte senderHostID) {
        clearReceivedVoteAndLocked();
        byte currentPosition = 0;
        
        while (currentPosition < data.length) {
            byte sender = data[currentPosition + Message.OFFSET_SENDER];
            if (sender == 0) {
                break;
            }
            
            if (sender == senderHostID) {
                if (data[currentPosition + Message.OFFSET_FLAGS] == Message.FLAG_NORMAL) {
                    System.arraycopy(data, currentPosition + Message.OFFSET_VOTE,
                            receivedVote, 0, hostCount);
                    
                    System.arraycopy(data, currentPosition + Message.OFFSET_VOTE + hostCount,
                            receivedLocked, 0, hostCount);
                
                    return true;
                    
                } else {
                    return false;
                }
            }
            
            currentPosition += normalMessageLength();
        }
        
        return false;
    }

    public Item getItem() {
        if (finished && shouldReturnReceivedItem) {
            byte[] description = new byte[receivedItemDescriptionLength];
            System.arraycopy(receivedItemDescription, 0, description, 0, description.length);
            
            byte[] content = new byte[receivedItemContentLength];
            System.arraycopy(receivedItemContent, 0, content, 0, content.length);
            
            return new Item(description, content);
                
        } else {
            return new Item();
        }
    }
    
    public boolean hasErrorOccured() {

        // not used yet
        return false;
    }
    
    public boolean isComputationFinished() {
        return computationFinished;
    }
    
    public boolean isFinished() {
        return finished;
    }
    
    public void start() {
        computationFinished = false;
        finished = false;
        
        round = 0;
        clearMessages();
        
        logger.info("Starting round " + round);
        
        maxFaultyProcesses = computeMaxFaultyProcesses();
        logger.debug("Maximum faulty processes (t) is: " + maxFaultyProcesses);
        
        if (isInitiator()) {
            logger.trace("I am the Initiator");
            
            byte k = pickRandomNumber();
            logger.debug("Generated random number k=" + k);
            
            fakeRoundsCount = k;
            
            prepareInitiatorMessages();
            
        } else {
            logger.trace("I am NOT the Initiator");
            
            prepareMessageWithItem();
        }
        
        computationFinished = true;
    }
    
    byte[] messagesToBytes(List messages) {
        clearData();

        byte currentPosition = 0;
        
        for (Iterator i = messages.iterator(); i.hasNext(); ) {
            Message message = (Message) i.next();
            System.arraycopy(message.getContent(), 0, data, currentPosition, message.getSize());
            
            currentPosition += message.getSize();
        }
        
        return data;
    }
    
    private void terminateProtocolWithError() {
        shouldReturnReceivedItem = false;
        finished = true;
    }
    
    public void startNextRound(List messages) {
        if (finished) {
            throw new IllegalStateException("Protocol already finished");
        }
        
        computationFinished = false;
        round++;
        data = messagesToBytes(messages);
        clearMessages();
        
        logger.info("Starting round " + round);
        
        if (round == 1) {
            if (hasReceivedItemFromSourceHost()) {
                logger.debug("Received an item from source host");
                
                if (isReceivedItemCorrect()) {
                    logger.debug("Received item is correct");
                    
                } else {
                    logger.error("Received item is NOT correct");
                    
                    terminateProtocolWithError();
                    return;
                }
                
            } else {
                logger.error("Did not receive the item from source host");
                
                terminateProtocolWithError();
                return;
            }
            
            if (!isInitiator()) {
                if (hasReceivedRandomNumberFromInitiator()) {
                    logger.debug("Received fake rounds count: " + fakeRoundsCount);
                    
                } else {
                    logger.error("Did not receive fake rounds count");
                    
                    terminateProtocolWithError();
                    return;
                }
            }
            
            prepareFakeMessages();
            
        } else if (round <= fakeRoundsCount) {
            //logger.info("This round is just fake");
            
            if (hasReceivedFakeMessagesFromAllHosts()) {
                logger.info("Fake messages received from all hosts");
                
            } else {
                logger.error("Did not receive fake messages from all other hosts");
                terminateProtocolWithError();
                return;
            }
            
            prepareFakeMessages();
            
        } else if (round == fakeRoundsCount + 1) {
            if (hasReceivedFakeMessagesFromAllHosts()) {
                logger.info("Fake messages received from all hosts");
                
            } else {
                logger.error("Did not receive fake messages from all hosts");
                terminateProtocolWithError();
                return;
            }
            
            //do not prepare fake messages anymore
            logger.info("Real calculation is starting now");
            
            vote.add((byte) hostID);
            newVote.add((byte) hostID);
            locked.clear();
            suspected.clear();
            
            prepareNormalMessages();
            
        } else if (round <= maxFaultyProcesses + fakeRoundsCount + 2) {
            newVote.clear();
            
            byte[] voteArray = vote.toArray();
            byte[] newVoteArray = newVote.toArray();
            
            for (byte i = 1; i <= hostCount; i++) {

                // We do not send messages to ourselves, so ignore this case - is this a problem?
                if (i != hostID) {
                    if (!suspected.contains(i)) {
                        if (hasReceivedNormalMessageFromHost(i)) {
                            logger.debug("Received a normal messaged from host " + i);
                            
                            for (byte j = 0; j < hostCount; j++) {
                                if (receivedVote[j] == HostSet.TRUE && voteArray[j] == HostSet.FALSE) {
                                    voteArray[j] = HostSet.TRUE;
                                    newVoteArray[j] = HostSet.TRUE;
                                    locked.addAll(receivedLocked);
                                }
                            }
                        }
                        
                    } else if (!locked.contains(i)) {
                        logger.warn("Did not receive a normal message from host " + i);
                        
                        suspected.add(i);
                    }
                }
            }
            
            if (suspected.size() > maxFaultyProcesses) {
                logger.error("Suspecting already at least half of all hosts");
                
                terminateProtocolWithError();
                return;
            }

            if (!locked.contains((byte) hostID)) {
                
                if ((round - (fakeRoundsCount + 1) > suspected.size()) || !locked.isEmpty()) {
                    locked.add((byte) hostID);
                }
                
            } else if (locked.size() > maxFaultyProcesses) {
                decide();
                return;
            }
            
            if (round <= maxFaultyProcesses + fakeRoundsCount + 1) {
                prepareNormalMessages();
                
            } else {
                assert round == maxFaultyProcesses + fakeRoundsCount + 2;
                
                decide();
                return;
            }
            
        } else {
            assert false;
        }
        
        computationFinished = true;
    }
    
    private void decide() {
        shouldReturnReceivedItem = true;
        
        for (byte i = 1; i <= hostCount; i++) {
            if (!vote.contains(i)) {
                shouldReturnReceivedItem = false;
                break;
            }
        }

        finished = true;
    }
    
    private byte normalMessageLength() {
        return (byte) (2 * hostCount + 5);
    }
    
    /* Properties */
    
    public void setHostCount(byte hostCount) {
        this.hostCount = hostCount;
    }
    
    public void setHostID(byte hostID) {
        this.hostID = hostID;
    }
    
    public void setDestinationHost(byte destinationHost) {
        this.destinationHost = destinationHost;
    }
    
    public void setSourceHost(byte sourceHost) {
        this.sourceHost = sourceHost;
    }
    
    public void setOwnItem(Item ownItem) {
        this.ownItem = ownItem;
    }
    
    public void setWantedItemDescription(byte[] wantedItemDescription) {
        this.wantedItemDescription = wantedItemDescription;
        
        //TODO: temporary solution to overcome configuration conversion problems
        Utils.convertNumbers(this.wantedItemDescription);
    }
    
    /* Properties not to be set via IoC */
    
    void setMessages(byte[] messages) {
        this.messages = messages;
    }
    
    void setRound(byte round) {
        this.round = round;
    }
    
    void setFakeRoundsCount(byte fakeRoundsCount) {
        this.fakeRoundsCount = fakeRoundsCount;
    }
    
    void setLocked(byte[] locked) {
        this.locked.clear();
        
        this.locked.addAll(locked);
    }
    
    void setNewVote(byte[] newVote) {
        this.newVote.clear();
        
        this.newVote.addAll(newVote);
    }
    
    byte[] getData() {
        return data;
    }
    
    void setData(byte[] data) {
        this.data = data;
    }
    
    byte getFakeRoundsCount() {
        return fakeRoundsCount;
    }
    
    byte[] getReceivedItemContent() {
        return receivedItemContent;
    }
    
    byte getReceivedItemContentLength() {
        return receivedItemContentLength;
    }
    
    byte[] getReceivedItemDescription() {
        return receivedItemDescription;
    }
    
    byte getReceivedItemDescriptionLength() {
        return receivedItemDescriptionLength;
    }
    
    void setReceivedItemDescription(byte[] receivedItemDescription) {
        this.receivedItemDescription = receivedItemDescription;
    }
    
    void setReceivedItemDescriptionLength(byte receivedItemDescriptionLength) {
        this.receivedItemDescriptionLength = receivedItemDescriptionLength;
    }
    
    byte[] getReceivedLocked() {
        return receivedLocked;
    }
    
    byte[] getReceivedVote() {
        return receivedVote;
    }
}
