/*
 * @(#)SmartCardCSFairExchangeProtocol.java
 *
 * $Revision: 1.2 $
 * $Date: 2006/04/06 19:10:12 $
 *
 * Copyright (c) 2005-2006 RWTH Aachen University, Lehr- und Forschungsgebiet Informatik 4.
 * All rights reserved.
 */

package de.rwth.informatik.lufgi4.trustedpals.protocol;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import opencard.core.service.CardRequest;
import opencard.core.service.SmartCard;
import opencard.core.terminal.CardTerminal;
import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CardTerminalRegistry;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.ResponseAPDU;
import opencard.opt.util.PassThruCardService;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.rwth.informatik.lufgi4.trustedpals.messaging.Message;

/**
 * 
 * 
 * @author Milan Fort
 * @version $Revision: 1.2 $, $Date: 2006/04/06 19:10:12 $
 * 
 * @since 1.0
 */
public final class SmartCardCSFairExchangeProtocol implements Protocol {

    //  possible values are up to 42
    private static final byte MAX_SUPPORTED_HOSTS = 10;
    
    private static final short MAX_SUPPORTED_DATA_LENGTH =
        (MAX_SUPPORTED_HOSTS - 1) * 6 + 1;
    
    private static Log logger = LogFactory.getLog(SmartCardCSFairExchangeProtocol.class);
    
    private static final byte CONSENSUSS_FAIR_EXCHANGE_CLA = (byte) 0xC2;
    
    private static final byte START_INS = (byte) 0x40;
    
    private static final byte START_NEXT_ROUND_INS = (byte) 0x50;
    
    private static final byte FINISHED_QUERY_INS = (byte) 0x60;

    private static final byte GET_MESSAGES_INS = (byte) 0x70;
    
    private static final byte GET_OUTPUT_INS = (byte) 0x80;

    private static final byte P1 = (byte) 0x00;
    
    private static final byte P2 = (byte) 0x00;
    
    private static final byte EMPTY = (byte) 0x00;

    private static final byte OFFSET_CDATA = 5;

    // Applet ID: A0 00 00 00 62 03 01 0D 04 01
    // 00 A4 04 00 0A A0 00 00 00 62 03 01 0D 04 01
    private CommandAPDU selectConsensusSFairExchangeApplet = new CommandAPDU(new byte[] {
            (byte) 0x00, (byte) 0xA4, (byte) 0x04, (byte) 0x00, (byte) 0x0A,  (byte) 0xA0,
            (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x62, (byte) 0x03,  (byte) 0x01,
            (byte) 0x0D, (byte) 0x04, (byte) 0x01, (byte) 0x7F
    });
    
    private CommandAPDU startProtocolAPDU = new CommandAPDU(new byte[] {
            CONSENSUSS_FAIR_EXCHANGE_CLA, START_INS, P1, P2, (byte) 0x00});

    private CommandAPDU finishedQueryAPDU = new CommandAPDU(new byte[] {
            CONSENSUSS_FAIR_EXCHANGE_CLA, FINISHED_QUERY_INS, P1, P2, (byte) 0x01});

    private CommandAPDU getMessagesAPDU = new CommandAPDU(new byte[] {
            CONSENSUSS_FAIR_EXCHANGE_CLA, GET_MESSAGES_INS, P1, P2, (byte) 0xFF});

    private CommandAPDU getOutputAPDU = new CommandAPDU(new byte[] {
            CONSENSUSS_FAIR_EXCHANGE_CLA, GET_OUTPUT_INS, P1, P2, (byte) 0xFF});
    
    private volatile boolean computationFinished = false;
    
    private byte round = 0;
    
    private volatile boolean errorOccured = false; 
    
    private SmartCard smartCard = null;
    
    private PassThruCardService card = null;
    
    public SmartCardCSFairExchangeProtocol() {}
    
    public void start() {
        computationFinished = false;
        errorOccured = false;
        
        round = 0;
        logger.info("Starting round " + round);
        
        Runnable r = new Runnable() {
            public void run() {
                try {
                    logger.debug("Starting the OCF framework");
                    SmartCard.start();

                    /* Dummy code to remove nonexistent terminal references from the registry. */
                    CardTerminalRegistry ctr = CardTerminalRegistry.getRegistry();
                    Enumeration terminals = ctr.getCardTerminals();
                    ctr.remove((CardTerminal) terminals.nextElement());
                    ctr.remove((CardTerminal) terminals.nextElement());

                    logger.debug("Waiting for a smartcard");
                    CardRequest cardRequest = new CardRequest(CardRequest.ANYCARD, null, null);
                    smartCard = SmartCard.waitForCard(cardRequest);
                    logger.debug("Smartcard detected");
                    
                    card =
                        (PassThruCardService) smartCard.getCardService(PassThruCardService.class, true);
                    
                    ResponseAPDU response = card.sendCommandAPDU(selectConsensusSFairExchangeApplet);
                    logger.debug("Selecting ConsensusS Fair Exchange Applet: 0x"
                            + Integer.toHexString(response.sw()));
                    
                    if (!normalResponse(response)) {
                        logger.error("Incorrect response received");
                        
                        errorOccured = true;
                        return;
                    }
                    
                    response = card.sendCommandAPDU(startProtocolAPDU);
                    logger.debug("Starting computation on the card: 0x"
                            + Integer.toHexString(response.sw()));

                    if (!normalResponse(response)) {
                        logger.error("Incorrect response received");
                        
                        errorOccured = true;
                        return;
                    }
                    
                } catch (Exception e) {
                    errorOccured = true;
                    
                    try {
                        SmartCard.shutdown();
                        
                    } catch (CardTerminalException cte) {
                        // ignore...
                    }
                    
                    throw new RuntimeException(e);
                    
                } finally {
                    computationFinished = true;    
                }
            }
        };
        
        Thread thread = new Thread(r, "Protocol Starter");
        thread.start();
    }
    
    private boolean normalResponse(ResponseAPDU response) {
        return (response.sw1() & 0xFF) == 0x90 && response.sw2() == 0x00;
    }
    
    public List getMessages() {
        if (!computationFinished) {
            throw new IllegalStateException("Computation has not finished yet");
        }
        
        try {
            ResponseAPDU response = card.sendCommandAPDU(getMessagesAPDU);
            logger.debug("Sending get messages request: 0x" + Integer.toHexString(response.sw()));
            
            byte[] messages = response.data();
            logger.debug("Retrieved data: " + ArrayUtils.toString(messages));
            
            List parsedMessages = new ArrayList();
            byte currentPosition = 0;

            while (currentPosition < messages.length) {
                byte sender = messages[currentPosition + Message.OFFSET_SENDER];
                if (sender == 0) {
                    break;
                }
                
//                byte hostCount = messages[currentPosition + Message.OFFSET_HOST_COUNT];
                byte round = messages[currentPosition + Message.OFFSET_ROUND];

                byte[] content = null;
                if (round == 0) {
                    byte descriptionLength = messages[currentPosition + Message.OFFSET_ITEM];
                    
                    if (descriptionLength == 0) {
                        content = new byte[6];
                        
                    } else {
                        byte itemContentLength = messages[currentPosition + Message.OFFSET_ITEM
                                                         + descriptionLength + 1];
                        
                        content = new byte[descriptionLength + itemContentLength + 2 + 5];
                    }
                    
                } else {
//                    assert hostCount == this.hostCount;
                    
                    content = new byte[normalMessageLength()];
                }
                
                System.arraycopy(messages, currentPosition, content, 0, content.length);
                Message message = new Message(content);
                parsedMessages.add(message);
                currentPosition += content.length;
            }
            
            return parsedMessages;
            
        } catch (Exception e) {
            errorOccured = true;
            
            logger.error("Error in communication with the smartcard", e);
            return new ArrayList();
        }
    }

    public Item getItem() {
        try {
            ResponseAPDU response = card.sendCommandAPDU(getOutputAPDU);
            logger.debug("Sending get output request: 0x" + Integer.toHexString(response.sw()));
            
            byte[] data = response.data();
            logger.debug("Retrieved data: " + ArrayUtils.toString(data));

            if (data.length < 4) {
                return new Item();
            }
            
            int descriptionLength = data[0];
            byte[] description = new byte[descriptionLength];
            System.arraycopy(data, 1, description, 0, description.length);
            
            int contentLength = data[description.length + 1];
            byte[] content = new byte[contentLength];
            System.arraycopy(data, description.length + 2, content, 0, content.length);
            
            Item item = new Item(description, content);

            return item;

            //SmartCard.shutdown();
            
        } catch (Exception e) {
            errorOccured = true;
            
            logger.error("Error in communication with the smartcard", e);
            return new Item();
        }
    }
    
    public boolean hasErrorOccured() {
         return errorOccured;
    }
    
    public boolean isComputationFinished() {
        return computationFinished;
    }
    
    public boolean isFinished() {
        if (computationFinished) {
            try {
                ResponseAPDU response = card.sendCommandAPDU(finishedQueryAPDU);
                logger.debug("Sending finished query: 0x" + Integer.toHexString(response.sw()));

                byte[] data = response.data();
                logger.debug("Retrieved data: " + ArrayUtils.toString(data));
                
                return data.length == 1 ? data[0] == 1 : false;
                
            } catch (Exception e) {
                errorOccured = true;
                
                logger.error("Error in communication with the smartcard", e);
                return false;
            }
            
        } else {
            return false;    
        }
    }
    
    byte[] messagesToBytes(List messages) {
        byte[] temp = new byte[MAX_SUPPORTED_DATA_LENGTH];
        byte currentPosition = 0;
        
        for (Iterator i = messages.iterator(); i.hasNext(); ) {
            Message message = (Message) i.next();
            System.arraycopy(message.getContent(), 0, temp, currentPosition, message.getSize());
            
            currentPosition += message.getSize();
        }
        
        byte[] result = new byte[currentPosition];
        System.arraycopy(temp, 0, result, 0, result.length);
        
        return result;
    }
    
    public void startNextRound(final List messages) {
        computationFinished = false;
        round++;
        logger.info("Starting round " + round);
        
        Runnable r = new Runnable() {
            public void run() {
                try {
                    byte[] data = messagesToBytes(messages);
                    
//                    CommandAPDU command = new CommandAPDU(data.length + 5);
//                    byte[] buffer = command.getBuffer(); 
//                    buffer[0] = CONSENSUSS_FAIR_EXCHANGE_CLA;
//                    buffer[1] = START_NEXT_ROUND_INS;
//                    buffer[2] = P1;
//                    buffer[3] = P2;
//                    buffer[4] = (byte) data.length;
//                    
//                    System.arraycopy(data, 0, buffer, OFFSET_CDATA, data.length);
                    
                    CommandAPDU command = new CommandAPDU(new byte[] {
                            CONSENSUSS_FAIR_EXCHANGE_CLA, START_NEXT_ROUND_INS, P1, P2, (byte) data.length,
                            EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, 
                            EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
                            EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
                            EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY});
                    
                    byte[] buffer = command.getBuffer();
                    System.arraycopy(data, 0, buffer, OFFSET_CDATA, data.length);
                    
                    byte[] tempbuffer = command.getBuffer();
                    logger.debug("Preparing command APDU: " + ArrayUtils.toString(tempbuffer));
                    
                    ResponseAPDU response = card.sendCommandAPDU(command);
                    logger.debug("Sending start next round request: 0x"
                            + Integer.toHexString(response.sw()));
                    
                    if (!normalResponse(response)) {
                        logger.error("Incorrect response received");
                        
                        errorOccured = true;
                        return;
                    }
                    
                } catch (Exception e) {
                    errorOccured = true;
                    
                    try {
                        SmartCard.shutdown();
                        
                    } catch (CardTerminalException cte) {
                        // ignore...
                    }
                    
                    throw new RuntimeException(e);
                    
                } finally {
                    computationFinished = true;    
                }
            }
        };
        
        Thread thread = new Thread(r, "Round " + round + " Starter");
        thread.start();
    }
    
    private byte normalMessageLength() {
        return (byte) (1 + 5);
    }
}
