/*
 * @(#)Random.java
 *
 * $Revision: 1.3 $
 * $Date: 2006/04/06 19:10:12 $
 *
 * Copyright (c) 2005-2006 RWTH Aachen University, Lehr- und Forschungsgebiet Informatik 4.
 * All rights reserved.
 */

package de.rwth.informatik.lufgi4.trustedpals.random;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;

/**
 * A Pseudo-random number generator implemented using a secure hash function. 
 * 
 * @author Milan Fort
 * @version $Revision: 1.3 $, $Date: 2006/04/06 19:10:12 $
 * 
 * @since 1.0
 */
public final class Random {
    
    private static final short MAX_SEED_LENGTH = 10;
    
    private static final short HASH_LENGTH = 20;
    
    private short seedLength = 0;
    
    private short seedByteToUpdate = 0;
    
    private byte[] seed = null;
    
    private byte[] hash = null;

    public Random() {
        seedLength = 0;
        seedByteToUpdate = 0;
        seed = new byte[MAX_SEED_LENGTH];
        hash = new byte[HASH_LENGTH];
    }

    public void setSeed(byte[] buffer, short offset, short length) {
        Arrays.fill(seed, (byte) 0);
        
        if (length > MAX_SEED_LENGTH || length <= 0) {
            throw new IllegalArgumentException("Invalid seed length: " + length);
        }
        
        seedLength = length;
        System.arraycopy(buffer, offset, seed, 0, length);
        
        seedByteToUpdate = 0;
    }
    
    public byte nextRandomNumber() {
        hash = computeSHA1Digest(seed, (short) 0, seedLength);
            
        updateSeed();
        
        return hasOddOnesCount(hash) ? (byte) 1 : (byte) 0;
    }
    
    short getSeedLength() {
        return seedLength;
    }
    
    byte[] getSeed() {
        return seed;
    }

    short getSeedByteToUpdate() {
        return seedByteToUpdate;
    }
    
    void updateSeed() {
        seed[seedByteToUpdate] = increaseByte(seed[seedByteToUpdate]);
        
        increaseSeedByteToUpdate();
    }
    
    static byte increaseByte(byte x) {
        return (byte) (x == 127 ? -128 : x + 1);
    }
    
    short increaseSeedByteToUpdate() {
        assert seedLength > 0;
        
        seedByteToUpdate = (short) ((seedByteToUpdate + 1) % seedLength);
        
        return seedByteToUpdate;
    }
    
    static byte computeOnesCount(byte x) {
        byte onesCount = 0;

        for (short i = 1; i <= 128; i *= 2) {
            onesCount += (x & i) == i ? 1 : 0;
        }
        
//        onesCount += (x & 0x01) == 0x01 ? 1 : 0;
//        onesCount += (x & 0x02) == 0x02 ? 1 : 0;
//        onesCount += (x & 0x04) == 0x04 ? 1 : 0;
//        onesCount += (x & 0x08) == 0x08 ? 1 : 0;
//        onesCount += (x & 0x10) == 0x10 ? 1 : 0;
//        onesCount += (x & 0x20) == 0x20 ? 1 : 0;
//        onesCount += (x & 0x40) == 0x40 ? 1 : 0;
//        onesCount += (x & 0x80) == 0x80 ? 1 : 0;
        
        return onesCount;
    }
    
    static boolean hasOddOnesCount(byte[] x) {
        short onesCount = 0;
        
        for (short i = 0; i < x.length; i++) {
            onesCount += computeOnesCount(x[i]);
        }
        
        return onesCount % 2 == 1;
    }
    
    private byte[] computeMD5Digest(byte[] input, short offset, short length) {
        try {
            MessageDigest messageDigest = MessageDigest.getInstance("MD5");
            
            byte[] data = new byte[length];
            System.arraycopy(input, offset, data, 0, length);

            messageDigest.update(data);
            
            return messageDigest.digest();
            
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }
    
    private byte[] computeSHA1Digest(byte[] input, short offset, short length) {
        try {
            MessageDigest messageDigest = MessageDigest.getInstance("SHA1");

            byte[] data = new byte[length];
            System.arraycopy(input, offset, data, 0, length);
            
            messageDigest.update(data);
            
            return messageDigest.digest();
            
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }

    /** Self-testing module. */
    public static void main(String[] args) {
        Random random = new Random();
        random.setSeed(new byte[] {1, 2, 3}, (short) 0, (short) 3);
        
        int attempts = 100;
        int zeroCount = 0;
        System.out.println("Starting...");

        for (int i = 0; i < attempts; i++) {
            byte randomNumber = random.nextRandomNumber();
            
            if (randomNumber == 0) {
                zeroCount++;
            }
            
            System.out.println("Next random number is: " + randomNumber);
        }
        
        System.out.println("Finished.");
        System.out.printf("Zeros: %.2f%%", zeroCount / (attempts / 100f));
    }
}
